import unittest2
from timeseries import TimeDimension, TimeSlice
from duration import Seconds, Milliseconds
from samplerate import SR44100


class TimeDimensionTests(unittest2.TestCase):
    def test_equals(self):
        td1 = TimeDimension(Seconds(1), Milliseconds(900))
        td2 = TimeDimension(Seconds(1), Milliseconds(900))
        self.assertEqual(td1, td2)

    def test_not_equal(self):
        td1 = TimeDimension(Seconds(2), Milliseconds(900))
        td2 = TimeDimension(Seconds(1), Milliseconds(900))
        self.assertNotEqual(td1, td2)

    def test_raises_if_frequency_is_not_timedelta_instance(self):
        self.assertRaises(ValueError, lambda: TimeDimension('s'))

    def test_raises_if_duration_is_not_timedelta_instance(self):
        self.assertRaises(ValueError, lambda: TimeDimension(Seconds(2), 's'))

    def test_duration_is_equal_to_frequency_if_not_provided(self):
        td = TimeDimension(Seconds(1))
        self.assertEqual(Seconds(1), td.frequency)
        self.assertEqual(Seconds(1), td.duration)

    def test_integer_based_slice(self):
        td = TimeDimension(*SR44100(), size=44100 * 5)
        sl = td.integer_based_slice(TimeSlice(duration=Seconds(1)))
        self.assertEqual(slice(0, 44100), sl)


__version__ = '0.20.11'

from timeseries import \
    Hours, Minutes, Seconds, Milliseconds, Microseconds, Picoseconds, \
    SR11025, SR22050, SR44100, SR48000, SR96000, HalfLapped, Stride, \
    TimeSlice, VariableRateTimeSeries, VariableRateTimeSeriesFeature, \
    SampleRate, AudioSamples, TimeDimension, audio_sample_rate

from soundfile import \
    MetaData, AudioMetaDataEncoder, FreesoundOrgConfig, \
    OggVorbis, OggVorbisDecoder, OggVorbisEncoder, OggVorbisFeature, \
    OggVorbisWrapper, \
    AudioStream, \
    Resampler

from spectral import \
    SlidingWindow, OggVorbisWindowingFunc, WindowingFunc, \
    FFT, MDCT, DCT, DCTIV, BarkBands, Chroma, BFCC, SpectralCentroid, \
    SpectralFlatness, AWeighting, LinearScale, LogScale, FrequencyBand, \
    FrequencyScale, FrequencyDimension

from segment import \
    MeasureOfTransience, MovingAveragePeakPicker, SparseTimestampDecoder, \
    SparseTimestampEncoder, TimeSliceDecoder, TimeSliceFeature, ComplexDomain

from synthesize import \
    FFTSynthesizer, DCTSynthesizer, TickSynthesizer, NoiseSynthesizer, \
    SineSynthesizer, DCTIVSynthesizer, MDCTSynthesizer

from learn import \
    KMeans, BinaryRbm, LinearRbm, Learned, \
    MeanStdNormalization, UnitNorm, Log, Multiply, PreprocessingPipeline, \
    Slicer, Flatten, ReservoirSampler, TemplateMatch, simple_settings

from ui import ZoundsApp, ZoundsSearch, RangeUnitUnsupportedException

from index import \
    Contiguous, Offsets, HammingDistanceScorer, PackedHammingDistanceScorer, \
    ConstantRateTimeSliceBuilder, VariableRateTimeSliceBuilder, Search, \
    SearchResults, hamming_index

from basic import \
    Slice, Sum, Max, Pooled, process_dir, stft, audio_graph, with_onsets, resampled

from util import \
    simple_lmdb_settings, simple_in_memory_settings, \
    simple_object_storage_settings

from nputil import sliding_window

from core import IdentityDimension, ArrayWithUnits

from persistence import ArrayWithUnitsFeature, AudioSamplesFeature



from basedimension import BaseDimensionEncoder, BaseDimensionDecoder
from zounds.spectral import FrequencyBand, FrequencyDimension
import zounds.spectral


class FrequencyDimensionEncoder(BaseDimensionEncoder):
    def __init__(self):
        super(FrequencyDimensionEncoder, self).__init__(FrequencyDimension)

    def dict(self, o):
        return dict(
                n_bands=o.scale.n_bands,
                start_hz=o.scale.frequency_band.start_hz,
                stop_hz=o.scale.frequency_band.stop_hz,
                always_even=o.scale.always_even,
                name=o.scale.__class__.__name__)


class FrequencyDimensionDecoder(BaseDimensionDecoder):
    def __init__(self):
        super(FrequencyDimensionDecoder, self).__init__(FrequencyDimension)

    def args(self, d):
        band = FrequencyBand(d['start_hz'], d['stop_hz'])
        # KLUDGE: This assumes that all FrequencyScale-derived classes live in
        # the zounds.spectral module
        scale_class = getattr(zounds.spectral, d['name'])
        scale = scale_class(band, d['n_bands'], always_even=d['always_even'])
        return scale,



import unittest2
import featureflow
from random_samples import ReservoirSampler
from preprocess import \
    UnitNorm, MeanStdNormalization, PreprocessingPipeline, Pipeline
from learn import LinearRbm, Learned
from zounds.util import simple_in_memory_settings
import numpy as np


class Iterator(featureflow.Node):
    def __init__(self, needs=None):
        super(Iterator, self).__init__(needs=needs)

    def _process(self, data):
        for d in data:
            yield d


def build_classes():

    class Settings(featureflow.PersistenceSettings):
        _id = 'rbm'
        id_provider = featureflow.StaticIdProvider(_id)
        key_builder = featureflow.StringDelimitedKeyBuilder()
        database = featureflow.InMemoryDatabase(key_builder=key_builder)

    class Rbm(featureflow.BaseModel, Settings):
        iterator = featureflow.Feature(
                Iterator,
                store=False)

        shuffle = featureflow.NumpyFeature(
                ReservoirSampler,
                nsamples=1000,
                needs=iterator,
                store=True)

        unitnorm = featureflow.PickleFeature(
                UnitNorm,
                needs=shuffle,
                store=False)

        meanstd = featureflow.PickleFeature(
                MeanStdNormalization,
                needs=unitnorm,
                store=False)

        rbm = featureflow.PickleFeature(
                LinearRbm,
                hdim=64,
                epochs=5,
                needs=meanstd,
                store=False)

        pipeline = featureflow.PickleFeature(
                PreprocessingPipeline,
                needs=(unitnorm, meanstd, rbm),
                store=True)

    return Rbm


def data():
    for i in xrange(100):
        yield np.random.random_sample((np.random.randint(10, 100), 3))


class RbmTests(unittest2.TestCase):
    def test_can_retrieve_rbm_pipeline(self):
        Rbm = build_classes()
        Rbm.process(iterator=data())
        self.assertIsInstance(Rbm().pipeline, Pipeline)


class LearnedTests(unittest2.TestCase):
    def test_can_use_learned_feature(self):
        Rbm = build_classes()
        Rbm.process(iterator=data())
        l = Learned(learned=Rbm())
        results = list(l._process(np.random.random_sample((33, 3))))[0]
        self.assertEqual((33, 64), results.shape)

    def test_pipeline_changes_version_when_recomputed(self):
        Rbm = build_classes()
        Rbm.process(iterator=data())
        v1 = Learned(learned=Rbm()).version
        v2 = Learned(learned=Rbm()).version
        self.assertEqual(v1, v2)
        Rbm.process(iterator=data())
        v3 = Learned(learned=Rbm()).version
        self.assertNotEqual(v1, v3)

    def test_pipeline_does_not_store_computed_data_from_training(self):
        Rbm = build_classes()
        Rbm.process(iterator=data())
        rbm = Rbm()
        pipeline_data = rbm.pipeline.processors[-1].data
        self.assertIsNone(pipeline_data)



import unittest2
import numpy as np
from featureflow import BaseModel, Node, PersistenceSettings
from basic import Merge
from zounds.timeseries import  Milliseconds, TimeDimension
from zounds.persistence import ArrayWithUnitsFeature
from zounds.core import ArrayWithUnits, IdentityDimension


class MergeTester(Node):
    def __init__(
            self,
            total_frames=100,
            increments_of=30,
            features=10,
            needs=None):
        super(MergeTester, self).__init__(needs=needs)
        self.features = features
        self.increments_of = increments_of
        self.total_frames = total_frames

    def _process(self, data):
        for i in xrange(0, self.total_frames, self.increments_of):
            size = min(self.increments_of, self.total_frames - i)
            td = TimeDimension(frequency=Milliseconds(500))
            yield ArrayWithUnits(
                    np.zeros((size, self.features)), [td, IdentityDimension()])


class MergeTests(unittest2.TestCase):
    def test_raises_if_single_source(self):
        class Document(BaseModel, PersistenceSettings):
            source = ArrayWithUnitsFeature(
                    MergeTester,
                    store=True)
            merged = ArrayWithUnitsFeature(
                    Merge,
                    needs=source,
                    store=True)

        self.assertRaises(ValueError, lambda: Document.process(source=''))

    def test_raises_if_single_element_iterable(self):
        class Document(BaseModel, PersistenceSettings):
            source = ArrayWithUnitsFeature(
                    MergeTester,
                    store=True)
            merged = ArrayWithUnitsFeature(
                    Merge,
                    needs=[source],
                    store=True)

        self.assertRaises(ValueError, lambda: Document.process(source=''))

    def test_can_combine_two_sources_at_same_rate(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    store=True)
            source2 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    store=True)
            merged = ArrayWithUnitsFeature(
                    Merge,
                    needs=[source1, source2],
                    store=True)

        _id = Document.process(source1='', source2='')
        doc = Document(_id)
        self.assertEqual((200, 20), doc.merged.shape)

    def test_can_combine_three_sources_at_same_rate(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    store=True)
            source2 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    store=True)
            source3 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    store=True)
            merged = ArrayWithUnitsFeature(
                    Merge,
                    needs=(source1, source2, source3),
                    store=True)

        _id = Document.process(source1='', source2='', source3='')
        doc = Document(_id)
        self.assertEqual((200, 30), doc.merged.shape)

    def test_can_combine_two_sources_at_different_rates(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=30,
                    store=True)
            source2 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=40,
                    store=True)
            merged = ArrayWithUnitsFeature(
                    Merge,
                    needs=(source1, source2),
                    store=True)

        _id = Document.process(source1='', source2='')
        doc = Document(_id)
        self.assertEqual((200, 20), doc.merged.shape)

    def test_can_combine_three_sources_at_different_rates(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=12,
                    store=True)
            source2 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=17,
                    store=True)
            source3 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=32,
                    store=True)
            merged = ArrayWithUnitsFeature(
                    Merge,
                    needs=(source1, source2, source3),
                    store=True)

        _id = Document.process(source1='', source2='', source3='')
        doc = Document(_id)
        self.assertEqual((200, 30), doc.merged.shape)

    def test_shortest_of_two_sources(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=190,
                    increments_of=30,
                    store=True)
            source2 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=40,
                    store=True)
            merged = ArrayWithUnitsFeature(
                    Merge,
                    needs=[source1, source2],
                    store=True)

        _id = Document.process(source1='', source2='')
        doc = Document(_id)
        self.assertEqual((190, 20), doc.merged.shape)

    def test_shortest_of_three_sources(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=12,
                    store=True)
            source2 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=185,
                    increments_of=17,
                    store=True)
            source3 = ArrayWithUnitsFeature(
                    MergeTester,
                    total_frames=50,
                    increments_of=32,
                    store=True)
            merged = ArrayWithUnitsFeature(
                    Merge,
                    needs=[source1, source2, source3],
                    store=True)

        _id = Document.process(source1='', source2='', source3='')
        doc = Document(_id)
        self.assertEqual((50, 30), doc.merged.shape)



from featureflow import Node, NotEnoughData
import marshal
import types
from collections import OrderedDict
import hashlib
import numpy as np


class Op(object):
    def __init__(self, func, **kwargs):
        super(Op, self).__init__()
        self._kwargs = kwargs
        self._func = marshal.dumps(func.func_code)
        self._version = self._compute_version()

    @property
    def version(self):
        return self._version

    def _compute_version(self):
        h = hashlib.md5(self._func)
        for v in self._kwargs.itervalues():
            try:
                h.update(v.version)
            except AttributeError:
                h.update(np.array(v))
        return h.hexdigest()

    def __getattr__(self, key):
        if key == '_kwargs':
            raise AttributeError()

        try:
            return self._kwargs[key]
        except KeyError:
            raise AttributeError(key)

    def __call__(self, arg, **kwargs):
        code = marshal.loads(self._func)
        f = types.FunctionType(
                code,
                globals(),
                'preprocess')
        kwargs.update(self._kwargs)
        return f(arg, **kwargs)


class PreprocessResult(object):
    def __init__(self, data, op, inversion_data=None, inverse=None, name=None):
        super(PreprocessResult, self).__init__()
        self.name = name
        self.inverse = inverse
        self.inversion_data = inversion_data
        self.data = data
        self.op = op

    def __str__(self):
        return 'PreprocessResult(name={name})'.format(**self.__dict__)

    def __repr__(self):
        return self.__str__()

    def __getattr__(self, key):
        if key == 'op':
            raise AttributeError()
        return getattr(self.op, key)

    def for_storage(self):
        return PreprocessResult(
                None,
                self.op,
                self.inversion_data,
                self.inverse,
                self.name)


class Preprocessor(Node):
    def __init__(self, needs=None):
        super(Preprocessor, self).__init__(needs=needs)

    def _forward_func(self):
        """
        Return a function that represents this processor's
        forward transform
        """

        def x(data, example_arg=None, another_one=None):
            return data

        return x

    def transform(self, **kwargs):
        return Op(self._forward_func(), **kwargs)

    def _inversion_data(self):
        """
        Return a function that computes any data needed for
        an inverse transform, if one is possible.  Otherwise,
        raise NotImplemented
        """

        def x(data, **kwargs):
            return kwargs

        return x

    def inversion_data(self, **kwargs):
        return Op(self._inversion_data(), **kwargs)

    def _backward_func(self):
        """
        Return a function that computes this processor's
        inverse transform, if one is possible.  Otherwise,
        raise NotImplemented
        """

        def x(data, **inversion_args):
            return data

        return x

    def inverse_transform(self):
        return Op(self._backward_func())

    def _extract_data(self, data):
        if isinstance(data, PreprocessResult):
            return data.data
        return data


class UnitNorm(Preprocessor):
    def __init__(self, needs=None):
        super(UnitNorm, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d):
            from zounds.nputil import safe_unit_norm
            from zounds.core import ArrayWithUnits
            normed = safe_unit_norm(d.reshape((d.shape[0], -1)))
            try:
                return ArrayWithUnits(normed, d.dimensions)
            except AttributeError:
                return normed

        return x

    def _inversion_data(self):
        def x(d):
            import numpy as np
            return dict(norm=np.linalg.norm(d, axis=1))

        return x

    def _backward_func(self):
        def x(d, norm=None):
            return d * norm[:, None]

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform()
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='UnitNorm')


class Log(Preprocessor):
    def __init__(self, needs=None):
        super(Log, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d):
            import numpy as np
            pos = np.abs(d)
            pos[pos == 0] = 1e-12
            return np.log(pos)

        return x

    def _inversion_data(self):
        def x(d):
            import numpy as np
            return dict(sign=np.sign(d))

        return x

    def _backward_func(self):
        def x(d, sign=None):
            import numpy as np
            return np.exp(d) * sign

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform()
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='Log')


class Slicer(Preprocessor):
    def __init__(self, slicex=None, needs=None):
        super(Slicer, self).__init__(needs=needs)
        self.fill_func = np.zeros
        self.slicex = slicex

    def _forward_func(self):
        def x(d, slicex=None):
            return d[..., slicex]

        return x

    def _inversion_data(self):
        def y(d, slicex=None, fill_func=None):

            try:
                ka = d.kwargs()

                def ff(shape):
                    return d.__class__(fill_func(shape), **ka)
            except AttributeError:

                def ff(shape):
                    return fill_func(shape)

            return dict(shape=d.shape, slicex=slicex, fill_func=ff)

        return y

    def _backward_func(self):
        def z(d, shape=None, fill_func=None, slicex=None):
            new_shape = d.shape[:1] + shape[1:]
            new_arr = fill_func(new_shape)
            new_arr[..., slicex] = d
            return new_arr

        return z

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(slicex=self.slicex)
        inv_data = self.inversion_data(
                fill_func=self.fill_func, slicex=self.slicex)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='Slicer')


class Flatten(Preprocessor):
    def __init__(self, needs=None):
        super(Flatten, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d):
            return d.reshape((d.shape[0], -1))

        return x

    def _inversion_data(self):
        def x(d):
            return dict(shape=d.shape)

        return x

    def _backward_func(self):
        def x(d, shape=None):
            return d.reshape(shape)

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform()
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='Flatten')


class MeanStdNormalization(Preprocessor):
    def __init__(self, needs=None):
        super(MeanStdNormalization, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d, mean=None, std=None):
            return (d - mean) / std

        return x

    def _backward_func(self):
        def x(d, mean=None, std=None):
            arr = d.copy()
            arr *= std
            arr += mean
            return arr

        return x

    def _process(self, data):
        data = self._extract_data(data)
        mean = data.mean(axis=0)
        std = data.std(axis=0)
        op = self.transform(mean=mean, std=std)
        inv_data = self.inversion_data(mean=mean, std=std)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='MeanStd')


class Multiply(Preprocessor):
    def __init__(self, factor=1, needs=None):
        super(Multiply, self).__init__(needs=needs)
        self.factor = factor

    def _forward_func(self):
        def x(d, factor=None):
            return d * factor

        return x

    def _backward_func(self):
        def x(d, factor=None):
            return d * (1.0 / factor)

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(factor=self.factor)
        inv_data = self.inversion_data(factor=self.factor)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='Multiply')


class Binarize(Preprocessor):
    def __init__(self, threshold=0.5, needs=None):
        super(Binarize, self).__init__(needs=needs)
        self.threshold = threshold

    def _forward_func(self):
        def x(d, threshold=None):
            import numpy as np
            data = np.zeros(d.shape, dtype=np.uint8)
            data[np.where(d > threshold)] = 1
            return data

        return x

    def _backward_func(self):
        def x(d):
            raise NotImplementedError()

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(threshold=self.threshold)
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='Binarize')


class Pipeline(object):
    def __init__(self, preprocess_results):
        self.processors = list(preprocess_results)
        self.version = hashlib.md5(
                ''.join([p.op.version for p in self.processors])).hexdigest()

    def __getitem__(self, index):
        return self.processors[index]

    def wrap_data(self, data):
        cls = data.__class__
        try:
            kwargs = data.kwargs()
        except AttributeError:
            kwargs = None
        return cls, kwargs

    def transform(self, data):
        inversion_data = []
        wrap_data = []
        for p in self.processors:
            inversion_data.append(p.inversion_data(data))
            wrap_data.append(self.wrap_data(data))
            data = p.op(data)
        return PipelineResult(data, self.processors, inversion_data, wrap_data)


class PipelineResult(object):
    def __init__(self, data, processors, inversion_data, wrap_data):
        super(PipelineResult, self).__init__()
        self.processors = processors[::-1]
        self.inversion_data = inversion_data[::-1]
        self.wrap_data = wrap_data[::-1]
        self.data = data

    def unwrap(self, data, wrap_data):
        cls, kwargs = wrap_data
        if kwargs is None:
            return data
        return cls(data, **kwargs)

    def inverse_transform(self):
        data = self.data
        for inv_data, wrap_data, p in \
                zip(self.inversion_data, self.wrap_data, self.processors):
            data = p.inverse(data, **inv_data)
            data = self.unwrap(data, wrap_data)
        return data


class PreprocessingPipeline(Node):
    def __init__(self, needs=None):
        super(PreprocessingPipeline, self).__init__(needs=needs)
        self._pipeline = OrderedDict((id(n), None) for n in needs)

    def _enqueue(self, data, pusher):
        self._pipeline[id(pusher)] = data

    def _dequeue(self):
        # TODO: Make this an aggregator
        if not self._finalized or not all(self._pipeline.itervalues()):
            raise NotEnoughData()

        return Pipeline(map(
                lambda x: x.for_storage(),
                self._pipeline.itervalues()))



import unittest2
from audiograph import resampled
from zounds.timeseries.samplerate import SR11025, SR22050
from zounds.timeseries.duration import Seconds
from zounds.util.persistence import simple_in_memory_settings
from zounds.synthesize.synthesize import NoiseSynthesizer


class ResampledTests(unittest2.TestCase):
    def test_audio_is_resampled(self):

        orig_sample_rate = SR22050()
        new_sample_rate = SR11025()

        rs = resampled(resample_to=new_sample_rate, store_resampled=True)

        @simple_in_memory_settings
        class Document(rs):
            pass

        synth = NoiseSynthesizer(orig_sample_rate)
        samples = synth.synthesize(Seconds(3))
        _id = Document.process(meta=samples.encode())
        doc = Document(_id)
        self.assertEqual(new_sample_rate, doc.resampled.samplerate)
        self.assertEqual(len(samples) // 2, len(doc.resampled))



from __future__ import division
import unittest2
import featureflow as ff
import numpy as np

from zounds.util import simple_in_memory_settings
from zounds.basic import stft
from zounds.timeseries import HalfLapped, Stride, SR44100, Seconds
from zounds.spectral import SlidingWindow
from zounds.synthesize import TickSynthesizer
from onset import \
    MeasureOfTransience, MovingAveragePeakPicker, TimeSliceFeature, \
    ComplexDomain
from zounds.persistence import ArrayWithUnitsFeature


class OnsetTests(unittest2.TestCase):

    def setUp(self):
        self.samplerate = SR44100()
        self.wscheme = HalfLapped()
        self.STFT = stft(
                store_fft=True,
                resample_to=self.samplerate,
                wscheme=self.wscheme)

    def ticks(self, samplerate, duration, tick_frequency):
        synth = TickSynthesizer(samplerate)
        samples = synth.synthesize(duration, tick_frequency)
        return samples.encode()

    def do_assertions(self, onset_class, feature_func):
        raw = self.ticks(self.samplerate, Seconds(4), Seconds(1))
        _id = onset_class.process(meta=raw)
        doc = onset_class(_id)
        slices = list(doc.slices)
        self.assertEqual(4, len(slices))
        frame_hop = self.wscheme.frequency

        self.assertLess(abs(Seconds(0) - slices[0].start), frame_hop)
        self.assertLess(abs(Seconds(1) - slices[1].start), frame_hop)
        self.assertLess(abs(Seconds(2) - slices[2].start), frame_hop)
        self.assertLess(abs(Seconds(3) - slices[3].start), frame_hop)

        self.assertLess(abs(Seconds(1) - slices[0].duration), frame_hop)
        self.assertLess(abs(Seconds(1) - slices[1].duration), frame_hop)
        self.assertLess(abs(Seconds(1) - slices[2].duration), frame_hop)
        # BUG: The last position reported by BasePeakPicker isn't guaranteed
        # to be the end of the file
        # self.assertLess(abs(Seconds(1) - slices[3].duration), frame_hop)
        return doc

    @unittest2.skip
    def test_complex_domain_onset_positions(self):
        class Settings(ff.PersistenceSettings):
            id_provider = ff.UuidProvider()
            key_builder = ff.StringDelimitedKeyBuilder()
            database = ff.InMemoryDatabase(key_builder=key_builder)

        class WithOnsets(self.STFT, Settings):
            onset_prep = ArrayWithUnitsFeature(
                    SlidingWindow,
                    needs=self.STFT.fft,
                    wscheme=self.wscheme * (1, 3),
                    store=False)

            complex_domain = ArrayWithUnitsFeature(
                    ComplexDomain,
                    needs=onset_prep,
                    store=False)

            sliding_detection = ArrayWithUnitsFeature(
                    SlidingWindow,
                    needs=complex_domain,
                    wscheme=self.wscheme * (1, 11),
                    padwith=5,
                    store=False)

            slices = TimeSliceFeature(
                    MovingAveragePeakPicker,
                    needs=sliding_detection,
                    aggregate=np.median,
                    store=True)

        self.do_assertions(WithOnsets, lambda x: x.complex_domain)

    def test_percussive_onset_positions(self):

        @simple_in_memory_settings
        class WithOnsets(self.STFT):
            transience = ArrayWithUnitsFeature(
                    MeasureOfTransience,
                    needs=self.STFT.fft,
                    store=True)

            sliding_detection = ArrayWithUnitsFeature(
                    SlidingWindow,
                    needs=transience,
                    wscheme=self.wscheme * Stride(frequency=1, duration=10),
                    padwith=5,
                    store=False)

            slices = TimeSliceFeature(
                    MovingAveragePeakPicker,
                    needs=sliding_detection,
                    aggregate=np.median,
                    store=True)

        self.do_assertions(WithOnsets, lambda x: x.transience)



import unittest2
from featureparser import FeatureParser
from zounds.basic import stft
from zounds.util import simple_in_memory_settings
from zounds.synthesize import NoiseSynthesizer
from zounds.timeseries import SR44100, Seconds


class FeatureParserTests(unittest2.TestCase):

    def setUp(self):
        @simple_in_memory_settings
        class Document(stft(store_fft=True)):
            pass

        synth = NoiseSynthesizer(SR44100())
        audio = synth.synthesize(Seconds(2))

        _id = Document.process(meta=audio.encode())
        doc = Document(_id)

        parser = FeatureParser(Document, locals())

        self.document = Document
        self.doc = doc
        self.parser = parser

    def test_can_extract_feature(self):
        parsed_doc, feature = self.parser.parse_feature('doc.fft')
        self.assertIs(parsed_doc, self.doc)
        self.assertIs(feature, self.document.features['fft'])

    def test_can_ignore_feature_in_larger_statement(self):
        parsed_doc, feature = self.parser.parse_feature('doc.fft.shape')
        self.assertIsNone(parsed_doc)
        self.assertIsNone(feature)

    def test_can_ignore_feauture_in_expression(self):
        parsed_doc, feature = self.parser.parse_feature('doc.fft *= 10')
        self.assertIsNone(parsed_doc)
        self.assertIsNone(feature)

    def test_can_ignore_feature_in_multiply_expression(self):
        parsed_doc, feature = self.parser.parse_feature('doc.fft * 10')
        self.assertIsNone(parsed_doc)
        self.assertIsNone(feature)



from collections import OrderedDict

import numpy as np
from featureflow import Node, NotEnoughData

from zounds.timeseries import VariableRateTimeSeries
from zounds.core import ArrayWithUnits


class Merge(Node):
    """
    Combine two or more sources into a single feature
    """

    def __init__(self, needs=None):
        super(Merge, self).__init__(needs=needs)
        exc_msg = 'you must supply two or more dependencies'
        try:
            needs = list(needs)
        except TypeError:
            raise ValueError(exc_msg)

        if len(needs) < 2:
            raise ValueError(exc_msg)

        self._cache = OrderedDict((id(n), None) for n in needs)

    def _enqueue(self, data, pusher):
        key = id(pusher)
        if self._cache[key] is None or self._cache[key].size == 0:
            self._cache[key] = data
        else:
            self._cache[key] = self._cache[key].concatenate(data)

    def _dequeue(self):
        if any(v is None or len(v) == 0 for v in self._cache.itervalues()):
            raise NotEnoughData()
        shortest = min(len(v) for v in self._cache.itervalues())
        output = OrderedDict(
                (k, v[:shortest]) for k, v in self._cache.iteritems())
        self._cache = OrderedDict(
                (k, v[shortest:]) for k, v in self._cache.iteritems())
        return output

    def _process(self, data):
        yield ArrayWithUnits.concat(data.values(), axis=1)


# KLUDGE: This implementation may currently only be used when consuming
# from a stored SparseTimeSliceFeature.  If it is run during initial computation,
# it will receive only timestamps, and the logic will break.  This is a bug
# in the BasePeakPicker class that needs to be fixed.  It should emit
# timeslices, and not timestamps
class Pooled(Node):
    def __init__(self, op=None, axis=None, needs=None):
        super(Pooled, self).__init__(needs=needs)
        self._timeslices = []
        self._timeseries = None
        self._op = op
        self._axis = axis

    def _enqueue(self, data, pusher):
        if isinstance(data, ArrayWithUnits):
            try:
                self._timeseries.concatenate(data)
            except AttributeError:
                self._timeseries = data
        else:
            self._timeslices.extend(data)

    def _dequeue(self):
        if not self._finalized:
            raise NotEnoughData()
        return self._timeslices, self._timeseries

    def _process(self, data):
        try:
            slices, series = data
            examples = [(ts, self._op(series[ts], axis=self._axis))
                        for ts in slices]
            yield VariableRateTimeSeries(examples)
        except Exception as e:
            print e
            raise


class Slice(Node):
    def __init__(self, sl=None, needs=None):
        super(Slice, self).__init__(needs=needs)
        self._sl = sl

    def _process(self, data):
        yield data[:, self._sl]


class Sum(Node):
    def __init__(self, axis=0, needs=None):
        super(Sum, self).__init__(needs=needs)
        self._axis = axis

    def _process(self, data):
        # TODO: This should be generalized.  Sum will have this same problem
        try:
            # data = np.sum(data, axis=self._axis)
            data = np.sum(axis=self._axis)
        except ValueError:
            print 'ERROR'
            data = data
        if data.shape[0]:
            yield data


class Max(Node):
    def __init__(self, axis=0, needs=None):
        super(Max, self).__init__(needs=needs)
        self._axis = axis

    def _process(self, data):
        # TODO: This should be generalized.  Sum will have this same problem
        try:
            # data = np.max(data, axis=self._axis)
            data = data.sum(axis=self._axis)
        except ValueError:
            print 'ERROR'
            data = data
        if data.shape[0]:
            yield data



import numpy as np
from featureflow import BaseModel, JSONFeature, ByteStream, ByteStreamFeature
from zounds.soundfile import \
    MetaData, AudioMetaDataEncoder, OggVorbis, OggVorbisFeature, AudioStream, \
    Resampler
from zounds.segment import \
    ComplexDomain, MovingAveragePeakPicker, TimeSliceFeature
from zounds.persistence import ArrayWithUnitsFeature, AudioSamplesFeature
from zounds.timeseries import SR44100, HalfLapped, Stride
from zounds.spectral import \
    SlidingWindow, OggVorbisWindowingFunc, FFT, BarkBands, SpectralCentroid, \
    Chroma, BFCC, DCT


def resampled(
        chunksize_bytes=2 * 44100 * 30 * 2,
        resample_to=SR44100(),
        store_resampled=False):
    """
    Create a basic processing pipeline that can resample all incoming audio
    to a normalized sampling rate for downstream processing, and store a
    convenient, compressed version for playback

    :param chunksize_bytes: The number of bytes from the raw stream to process
    at once
    :param resample_to: The new, normalized sampling rate
    :return: A simple processing pipeline
    """
    class Resampled(BaseModel):
        meta = JSONFeature(
                MetaData,
                store=True,
                encoder=AudioMetaDataEncoder)

        raw = ByteStreamFeature(
                ByteStream,
                chunksize=chunksize_bytes,
                needs=meta,
                store=False)

        ogg = OggVorbisFeature(
                OggVorbis,
                needs=raw,
                store=True)

        pcm = AudioSamplesFeature(
                AudioStream,
                needs=raw,
                store=False)

        resampled = AudioSamplesFeature(
                Resampler,
                needs=pcm,
                samplerate=resample_to,
                store=store_resampled)

    return Resampled


def stft(
        chunksize_bytes=2 * 44100 * 30 * 2,
        resample_to=SR44100(),
        wscheme=HalfLapped(),
        store_fft=False):
    class ShortTimeFourierTransform(BaseModel):
        meta = JSONFeature(
                MetaData,
                store=True,
                encoder=AudioMetaDataEncoder)

        raw = ByteStreamFeature(
                ByteStream,
                chunksize=chunksize_bytes,
                needs=meta,
                store=False)

        ogg = OggVorbisFeature(
                OggVorbis,
                needs=raw,
                store=True)

        pcm = AudioSamplesFeature(
                AudioStream,
                needs=raw,
                store=False)

        resampled = AudioSamplesFeature(
                Resampler,
                needs=pcm,
                samplerate=resample_to,
                store=False)

        windowed = ArrayWithUnitsFeature(
                SlidingWindow,
                needs=resampled,
                wscheme=wscheme,
                wfunc=OggVorbisWindowingFunc(),
                store=False)

        fft = ArrayWithUnitsFeature(
                FFT,
                needs=windowed,
                store=store_fft)

    return ShortTimeFourierTransform


def audio_graph(
        chunksize_bytes=2 * 44100 * 30 * 2,
        resample_to=SR44100(),
        freesound_api_key=None,
        store_fft=False):
    """
    Produce a base class suitable as a starting point for many audio processing
    pipelines.  This class resamples all audio to a common sampling rate, and
    produces a bark band spectrogram from overlapping short-time fourier
    transform frames.  It also compresses the audio into ogg vorbis format for
    compact storage.
    """

    class AudioGraph(BaseModel):
        meta = JSONFeature(
                MetaData,
                store=True,
                encoder=AudioMetaDataEncoder)

        raw = ByteStreamFeature(
                ByteStream,
                chunksize=chunksize_bytes,
                needs=meta,
                store=False)

        ogg = OggVorbisFeature(
                OggVorbis,
                needs=raw,
                store=True)

        pcm = AudioSamplesFeature(
                AudioStream,
                needs=raw,
                store=False)

        resampled = AudioSamplesFeature(
                Resampler,
                needs=pcm,
                samplerate=resample_to,
                store=False)

        windowed = ArrayWithUnitsFeature(
                SlidingWindow,
                needs=resampled,
                wscheme=HalfLapped(),
                wfunc=OggVorbisWindowingFunc(),
                store=False)

        dct = ArrayWithUnitsFeature(
                DCT,
                needs=windowed,
                store=True)

        fft = ArrayWithUnitsFeature(
                FFT,
                needs=windowed,
                store=store_fft)

        bark = ArrayWithUnitsFeature(
                BarkBands,
                needs=fft,
                samplerate=resample_to,
                store=True)

        centroid = ArrayWithUnitsFeature(
                SpectralCentroid,
                needs=bark,
                store=True)

        chroma = ArrayWithUnitsFeature(
                Chroma,
                needs=fft,
                samplerate=resample_to,
                store=True)

        bfcc = ArrayWithUnitsFeature(
                BFCC,
                needs=fft,
                store=True)

    return AudioGraph


def with_onsets(fft_feature):
    """
    Produce a mixin class that extracts onsets
    :param fft_feature: The short-time fourier transform feature
    :return: A mixin class that extracts onsets
    """

    class Onsets(BaseModel):
        onset_prep = ArrayWithUnitsFeature(
                SlidingWindow,
                needs=fft_feature,
                wscheme=HalfLapped() * Stride(frequency=1, duration=3),
                store=False)

        complex_domain = ArrayWithUnitsFeature(
                ComplexDomain,
                needs=onset_prep,
                store=False)

        sliding_detection = ArrayWithUnitsFeature(
                SlidingWindow,
                needs=complex_domain,
                wscheme=HalfLapped() * Stride(frequency=1, duration=11),
                padwith=5,
                store=False)

        slices = TimeSliceFeature(
                MovingAveragePeakPicker,
                needs=sliding_detection,
                aggregate=np.median,
                store=True)

    return Onsets



import unittest2
import numpy as np
from duration import Seconds
from samplerate import SR44100, SR11025, SampleRate
from zounds.timeseries import TimeDimension
from zounds.core import IdentityDimension
from audiosamples import AudioSamples


class AudioSamplesTest(unittest2.TestCase):

    def test_raises_if_not_audio_samplerate(self):
        arr = np.zeros(int(44100 * 2.5))
        one = Seconds(1)
        self.assertRaises(
                TypeError, lambda: AudioSamples(arr, SampleRate(one, one)))

    def test_raises_if_array_is_more_than_2d(self):
        arr = np.zeros((int(44100 * 2.5), 2, 2))
        self.assertRaises(
            ValueError, lambda:  AudioSamples(arr, SR44100()))

    def test_can_create_instance(self):
        arr = np.zeros(int(44100 * 2.5))
        instance = AudioSamples(arr, SR44100())
        self.assertIsInstance(instance, AudioSamples)
        length_seconds = instance.end / Seconds(1)
        self.assertAlmostEqual(2.5, length_seconds, places=6)

    def test_can_mix_two_instances(self):
        arr = np.ones(int(44100 * 2.5))
        first = AudioSamples(arr, SR44100())
        second = AudioSamples(arr, SR44100())
        mixed = first + second
        self.assertIsInstance(mixed, AudioSamples)
        self.assertEqual(SR44100(), mixed.samplerate)
        np.testing.assert_allclose(mixed, 2)

    def test_cannot_mix_two_instances_with_different_sample_rates(self):
        arr = np.ones(int(44100 * 2.5))
        first = AudioSamples(arr, SR44100())
        second = AudioSamples(arr, SR11025())
        self.assertRaises(ValueError, lambda: first + second)

    def test_can_add_plain_numpy_array(self):
        arr = np.ones(int(44100 * 2.5))
        first = AudioSamples(arr, SR44100())
        second = arr.copy()
        mixed = first + second
        self.assertIsInstance(mixed, AudioSamples)
        self.assertEqual(SR44100(), mixed.samplerate)
        np.testing.assert_allclose(mixed, 2)

    def test_channels_returns_one_for_one_dimensional_array(self):
        arr = np.zeros(int(44100 * 2.5))
        instance = AudioSamples(arr, SR44100())
        self.assertEqual(1, instance.channels)

    def test_channels_returns_two_for_two_dimensional_array(self):
        arr = np.zeros(int(44100 * 2.5))
        arr = np.column_stack((arr, arr))
        instance = AudioSamples(arr, SR44100())
        self.assertEqual(2, instance.channels)

    def test_samplerate_returns_correct_value(self):
        arr = np.zeros(int(44100 * 2.5))
        instance = AudioSamples(arr, SR44100())
        self.assertIsInstance(instance.samplerate, SR44100)

    def test_can_sum_to_mono(self):
        arr = np.zeros(int(44100 * 2.5))
        arr = np.column_stack((arr, arr))
        instance = AudioSamples(arr, SR44100())
        mono = instance.mono
        self.assertEqual(1, mono.channels)
        self.assertIsInstance(mono.samplerate, SR44100)

    def test_class_concat_returns_audio_samples(self):
        s1 = AudioSamples(np.zeros(44100 * 2), SR44100())
        s2 = AudioSamples(np.zeros(44100), SR44100())
        s3 = AudioSamples.concat([s1, s2])
        self.assertIsInstance(s3, AudioSamples)
        self.assertEqual(44100 * 3, len(s3))

    def test_instance_concat_returns_audio_samples(self):
        s1 = AudioSamples(np.zeros(44100 * 2), SR44100())
        s2 = AudioSamples(np.zeros(44100), SR44100())
        s3 = s1.concat([s1, s2])
        self.assertIsInstance(s3, AudioSamples)
        self.assertEqual(44100 * 3, len(s3))

    def test_concat_raises_for_different_sample_rates(self):
        s1 = AudioSamples(np.zeros(44100 * 2), SR44100())
        s2 = AudioSamples(np.zeros(44100), SR11025())
        self.assertRaises(ValueError, lambda: AudioSamples.concat([s1, s2]))

    def test_sum_along_time_axis(self):
        arr = np.zeros(int(44100 * 2.5))
        arr = np.column_stack((arr, arr))
        ts = AudioSamples(arr, SR44100())
        result = ts.sum(axis=0)
        self.assertIsInstance(result, np.ndarray)
        self.assertNotIsInstance(result, AudioSamples)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], IdentityDimension)

    def test_sum_along_second_axis(self):
        arr = np.zeros(int(44100 * 2.5))
        arr = np.column_stack((arr, arr))
        ts = AudioSamples(arr, SR44100())
        result = ts.sum(axis=1)
        self.assertIsInstance(result, AudioSamples)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], TimeDimension)



import os


def iter_files(base_path):
    for fn in os.listdir(base_path):
        yield os.path.join(base_path, fn)


def process_dir(base_path, process_func):
    for fp in iter_files(base_path):
        process_func(fp)


from __future__ import division
import unittest2
from frequencyscale import FrequencyBand, LinearScale, LogScale
from zounds.timeseries import SR44100
import numpy as np


class FrequencyBandTests(unittest2.TestCase):
    def test_can_create_from_center_frequency(self):
        fb = FrequencyBand.from_center(1000, 50)
        self.assertEqual(FrequencyBand(975, 1025), fb)

    def test_does_not_equal_non_frequency_band_class(self):
        fb = FrequencyBand(100, 200)
        self.assertNotEqual(fb, 10)

    def test_cannot_create_with_invalid_interval(self):
        self.assertRaises(ValueError, lambda: FrequencyBand(200, 100))


class FrequencyScaleTests(unittest2.TestCase):

    def test_can_get_all_even_sized_bands(self):
        samplerate = SR44100()
        scale = LinearScale.from_sample_rate(
                samplerate, 44100, always_even=True)
        log_scale = LogScale(FrequencyBand(20, 20000), 64)
        slices = [scale.get_slice(band) for band in log_scale]
        sizes = [s.stop - s.start for s in slices]
        self.assertTrue(
                not any([s % 2 for s in sizes]),
                'All slice sizes should be even but were {sizes}'
                    .format(**locals()))

    def test_can_get_single_band(self):
        fb1 = FrequencyBand(20, 20000)
        scale1 = LinearScale(fb1, 100)
        fb2 = scale1[10]
        self.assertIsInstance(fb2, FrequencyBand)

    def test_can_get_sub_scale(self):
        fb1 = FrequencyBand(20, 20000)
        scale1 = LinearScale(fb1, 100)
        scale2 = scale1[10:20]
        self.assertIsInstance(scale2, LinearScale)
        self.assertEqual(10, scale2.n_bands)

    def test_equals(self):
        fb1 = FrequencyBand(20, 20000)
        scale1 = LinearScale(fb1, 100)

        fb2 = FrequencyBand(20, 20000)
        scale2 = LinearScale(fb2, 100)

        self.assertEqual(scale1, scale2)

    def test_not_equal_when_scale_differs(self):
        fb1 = FrequencyBand(20, 20000)
        scale1 = LinearScale(fb1, 100)

        fb2 = FrequencyBand(20, 20000)
        scale2 = LogScale(fb2, 100)

        self.assertNotEqual(scale1, scale2)

    def test_not_equal_when_span_differs(self):
        fb1 = FrequencyBand(20, 20000)
        scale1 = LinearScale(fb1, 100)

        fb2 = FrequencyBand(20, 10000)
        scale2 = LinearScale(fb2, 100)

        self.assertNotEqual(scale1, scale2)

    def test_not_equal_when_bands_differ(self):
        fb1 = FrequencyBand(20, 20000)
        scale1 = LinearScale(fb1, 100)

        fb2 = FrequencyBand(20, 20000)
        scale2 = LinearScale(fb2, 50)

        self.assertNotEqual(scale1, scale2)


class LinearScaleTests(unittest2.TestCase):
    def test_matches_fftfreq(self):
        samplerate = SR44100()
        n_bands = 2048
        fft_freqs = np.fft.rfftfreq(n_bands, 1 / int(samplerate))
        bands = LinearScale.from_sample_rate(samplerate, n_bands // 2)
        linear_freqs = np.array([b.start_hz for b in bands])
        np.testing.assert_allclose(linear_freqs, fft_freqs[:-1], rtol=1e-3)

    def test_constant_bandwidth(self):
        scale = LinearScale(FrequencyBand(0, 22050), 1024)
        # taking the second-order differential should result in all zeros
        # if the bandwidths are a constant size
        diff = np.diff(list(scale.center_frequencies), n=2)
        print diff.min(), diff.max()
        np.testing.assert_allclose(diff, np.zeros(len(diff)), atol=1e-11)

    def test_get_slice_on_boundary(self):
        scale = LinearScale(FrequencyBand(0, 1000), 100)
        sl = scale.get_slice(FrequencyBand(500, 700))
        self.assertEqual(slice(49, 70), sl)

    def test_get_slice_between_boundary(self):
        scale = LinearScale(FrequencyBand(0, 1000), 10)
        sl = scale.get_slice(FrequencyBand(495, 705))
        self.assertEqual(slice(4, 7), sl)


class LogScaleTests(unittest2.TestCase):
    def test_variable_bandwidth(self):
        scale = LogScale(FrequencyBand(20, 22050), 100)
        diff = np.diff(list(scale.bandwidths))
        # All differences should be positive, as bandwidth should be
        # monotonically increasing
        self.assertTrue(np.all(diff > 0))



from dimension import DimensionEncoder, DimensionDecoder
from arraywithunits import \
    ArrayWithUnitsFeature, PackedArrayWithUnitsFeature, ArrayWithUnitsEncoder, \
    PackedArrayWithUnitsEncoder
from audiosamples import AudioSamplesFeature



import numpy as np
from duration import Picoseconds, Seconds
from samplerate import SampleRate
from zounds.core import Dimension


class TimeSlice(object):
    def __init__(self, duration=None, start=None):
        super(TimeSlice, self).__init__()

        if duration is not None and not isinstance(duration, np.timedelta64):
            raise ValueError('duration must be of type {t} but was {t2}'.format(
                    t=np.timedelta64, t2=duration.__class__))

        if start is not None and not isinstance(start, np.timedelta64):
            raise ValueError('start must be of type {t} but was {t2}'.format(
                    t=np.timedelta64, t2=start.__class__))

        self.duration = duration
        self.start = start or Picoseconds(0)

    def __add__(self, other):
        return TimeSlice(self.duration, start=self.start + other)

    def __radd__(self, other):
        return self.__add__(other)

    @property
    def end(self):
        return self.start + self.duration

    def __lt__(self, other):
        try:
            return self.start.__lt__(other.start)
        except AttributeError:
            return self.start.__lt__(other)

    def __gt__(self, other):
        try:
            return self.start.__gt__(other.start)
        except AttributeError:
            return self.start.__get__(other)

    def __le__(self, other):
        try:
            return self.start.__le__(other.start)
        except AttributeError:
            return self.start.__le__(other)

    def __ge__(self, other):
        try:
            return self.start.__ge__(other.start)
        except AttributeError:
            return self.start.__ge__(other)

    def __and__(self, other):
        delta = max(
                Picoseconds(0),
                min(self.end, other.end) - max(self.start, other.start))
        return TimeSlice(delta)

    def __contains__(self, other):
        print other
        if isinstance(other, np.timedelta64):
            return self.start < other < self.end
        if isinstance(other, TimeSlice):
            return other.start > self.start and other.end < self.end
        raise ValueError

    def __eq__(self, other):
        return self.start == other.start and self.duration == other.duration

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        dur = self.duration / Seconds(1) if self.duration is not None else None
        return '{cls}(start = {start}, duration = {duration})'.format(
                cls=self.__class__.__name__,
                start=self.start / Seconds(1),
                duration=dur)

    def __str__(self):
        return self.__repr__()


class TimeDimension(Dimension):
    def __init__(self, frequency=None, duration=None, size=None):
        super(TimeDimension, self).__init__()
        self.size = size
        if not isinstance(frequency, np.timedelta64):
            raise ValueError('duration must be of type {t} but was {t2}'.format(
                    t=np.timedelta64, t2=frequency.__class__))

        if duration is not None and not isinstance(duration, np.timedelta64):
            raise ValueError('start must be of type {t} but was {t2}'.format(
                    t=np.timedelta64, t2=duration.__class__))
        self.duration = duration or frequency
        self.frequency = frequency

    def __str__(self):
        fs = self.frequency / Picoseconds(int(1e12))
        ds = self.duration / Picoseconds(int(1e12))
        return 'TimeDimension(f={fs}, d={ds})'.format(**locals())

    def __repr__(self):
        return self.__str__()

    @property
    def samplerate(self):
        return SampleRate(self.frequency, self.duration)

    @property
    def overlap(self):
        return self.samplerate.overlap

    @property
    def duration_in_seconds(self):
        return self.duration / Picoseconds(int(1e12))

    @property
    def samples_per_second(self):
        return int(Picoseconds(int(1e12)) / self.frequency)

    @property
    def span(self):
        overlap = self.duration - self.frequency
        return TimeSlice((self.size * self.frequency) + overlap)

    @property
    def end(self):
        return self.span.end

    @property
    def end_seconds(self):
        return self.end / Picoseconds(int(1e12))

    def modified_dimension(self, size, windowsize, stepsize=None):
        stepsize = stepsize or windowsize
        yield TimeDimension(
                self.frequency * stepsize,
                (self.frequency * windowsize) + self.overlap)
        yield self

    def metaslice(self, index, size):
        return TimeDimension(self.frequency, self.duration, size)

    def integer_based_slice(self, ts):
        if not isinstance(ts, TimeSlice):
            return ts

        diff = self.duration - self.frequency
        start_index = \
            max(0, np.floor((ts.start - diff) / self.frequency))
        end = self.end if ts.duration is None else ts.end

        # KLUDGE: This is basically arbitrary, but the motivation is that we'd
        # like to differentiate between cases where the slice
        # actually/intentionally overlaps a particular sample, and cases where
        # the slice overlaps the sample by a tiny amount, due to rounding or
        # lack of precision (e.g. Seconds(1) / SR44100().frequency).
        ratio = np.round(end / self.frequency, 2)

        stop_index = np.ceil(ratio)
        return slice(int(start_index), int(stop_index))

    def __eq__(self, other):
        return \
            self.frequency == other.frequency \
            and self.duration == other.duration



class Dimension(object):
    """
    Class representing one dimension of a numpy array, making custom slices
    (e.g., time spans or frequency bands) possible
    """

    def __init__(self):
        super(Dimension, self).__init__()

    def modified_dimension(self, size, windowsize, stepsize=None):
        raise NotImplementedError()

    def metaslice(self, index, size):
        return self

    def integer_based_slice(self, index):
        raise NotImplementedError()

    def validate(self, size):
        pass


class IdentityDimension(Dimension):
    def __init__(self):
        super(IdentityDimension, self).__init__()

    def modified_dimension(self, size, windowsize, stepsize=None):
        if windowsize == slice(None) or (size / windowsize) == 1:
            yield IdentityDimension()
        else:
            raise ValueError()

    def integer_based_slice(self, index):
        return index

    def __eq__(self, other):
        return self.__class__ == other.__class__



from __future__ import division

import numpy as np
from featureflow import Node
from scipy.fftpack import dct
from scipy.stats.mstats import gmean

from frequencyscale import LinearScale
from psychacoustics import Chroma as ChromaScale, Bark as BarkScale
from tfrepresentation import FrequencyDimension
from zounds.core import ArrayWithUnits, IdentityDimension
from zounds.nputil import safe_log
from zounds.timeseries import SR44100, audio_sample_rate


class FFT(Node):
    def __init__(self, needs=None, axis=-1):
        super(FFT, self).__init__(needs=needs)
        self._axis = axis

    def _process(self, data):
        transformed = np.fft.fft(data, axis=self._axis)
        sl = [slice(None) for _ in xrange(len(transformed.shape))]
        positive = data.shape[self._axis] // 2
        sl[self._axis] = slice(0, positive, None)

        transformed = transformed[sl]

        sr = audio_sample_rate(
                int(data.shape[1] / data.dimensions[0].duration_in_seconds))
        scale = LinearScale.from_sample_rate(sr, transformed.shape[-1])

        yield ArrayWithUnits(
                transformed, [data.dimensions[0], FrequencyDimension(scale)])


class DCT(Node):
    """
    Type II Discrete Cosine Transform
    """

    def __init__(self,  axis=-1, scale_always_even=False, needs=None):
        super(DCT, self).__init__(needs=needs)
        self.scale_always_even = scale_always_even
        self._axis = axis

    def _process(self, data):

        transformed = dct(data, norm='ortho', axis=self._axis)

        sr = audio_sample_rate(
                int(data.shape[1] / data.dimensions[0].duration_in_seconds))
        scale = LinearScale.from_sample_rate(
                sr, transformed.shape[-1], always_even=self.scale_always_even)

        yield ArrayWithUnits(
                transformed, [data.dimensions[0], FrequencyDimension(scale)])


class DCTIV(Node):
    """
    Type IV Discrete Cosine Transform.  This transform is its own inverse
    """

    def __init__(self, scale_always_even=False, needs=None):
        super(DCTIV, self).__init__(needs=needs)
        self.scale_always_even = scale_always_even

    def _process_raw(self, data):
        l = data.shape[1]
        tf = np.arange(0, l)
        z = np.zeros((len(data), l * 2))
        z[:, :l] = data * np.exp(-1j * np.pi * tf / 2 / l)
        z = np.fft.fft(z)[:, :l]
        raw = np.sqrt(2 / l) * \
              np.real(z * np.exp(-1j * np.pi * (tf + 0.5) / 2 / l))
        return raw

    def _process(self, data):
        raw = self._process_raw(data)
        sr = audio_sample_rate(
                int(data.shape[1] / data.dimensions[0].duration_in_seconds))
        scale = LinearScale.from_sample_rate(
                sr, data.shape[1], always_even=self.scale_always_even)
        yield ArrayWithUnits(
                raw, [data.dimensions[0], FrequencyDimension(scale)])


class MDCT(Node):
    """
    Modified Discrete Cosine Transform
    """

    def __init__(self, needs=None):
        super(MDCT, self).__init__(needs=needs)

    def _process_raw(self, data):
        l = data.shape[1] // 2
        t = np.arange(0, 2 * l)
        f = np.arange(0, l)
        cpi = -1j * np.pi
        a = data * np.exp(cpi * t / 2 / l)
        b = np.fft.fft(a)
        c = b[:, :l]
        transformed = np.sqrt(2 / l) * np.real(
                c * np.exp(cpi * (f + 0.5) * (l + 1) / 2 / l))
        return transformed

    def _process(self, data):
        transformed = self._process_raw(data)

        sr = audio_sample_rate(data.dimensions[1].samples_per_second)
        scale = LinearScale.from_sample_rate(sr, transformed.shape[1])

        yield ArrayWithUnits(
                transformed, [data.dimensions[0], FrequencyDimension(scale)])


# TODO: This constructor should not take a samplerate; that information should
# be encapsulated in the data that's passed in
class Chroma(Node):
    def __init__(
            self,
            needs=None,
            samplerate=SR44100(),
            nbins=12,
            a440=440.):
        super(Chroma, self).__init__(needs=needs)
        self._nbins = nbins
        self._a440 = a440
        self._samplerate = samplerate
        self._chroma_scale = None

    def _process(self, data):
        data = np.abs(data)
        if self._chroma_scale is None:
            self._chroma_scale = ChromaScale( \
                    self._samplerate.samples_per_second,
                    data.shape[1] * 2,
                    nbands=self._nbins)

        yield ArrayWithUnits(
                self._chroma_scale.transform(data),
                [data.dimensions[0], IdentityDimension()])


# TODO: This constructor should not take a samplerate; that information should
# be encapsulated in the data that's passed in
class BarkBands(Node):
    def __init__(
            self,
            needs=None,
            samplerate=SR44100(),
            n_bands=100,
            start_freq_hz=50,
            stop_freq_hz=2e4):
        super(BarkBands, self).__init__(needs=needs)

        self._samplerate = samplerate
        self._n_bands = n_bands
        self._start_freq_hz = start_freq_hz
        self._stop_freq_hz = stop_freq_hz
        self._bark_scale = None

    def _process(self, data):
        data = np.abs(data)
        if self._bark_scale is None:
            self._bark_scale = BarkScale(
                    self._samplerate.samples_per_second,
                    data.shape[1] * 2,
                    self._n_bands,
                    self._start_freq_hz,
                    self._stop_freq_hz)

        yield ArrayWithUnits(
                self._bark_scale.transform(data),
                [data.dimensions[0], IdentityDimension()])


class SpectralCentroid(Node):
    """
    Indicates where the "center of mass" of the spectrum is. Perceptually,
    it has a robust connection with the impression of "brightness" of a
    sound.  It is calculated as the weighted mean of the frequencies
    present in the signal, determined using a Fourier transform, with
    their magnitudes as the weights...

    -- http://en.wikipedia.org/wiki/Spectral_centroid
    """

    def __init__(self, needs=None):
        super(SpectralCentroid, self).__init__(needs=needs)

    def _first_chunk(self, data):
        self._bins = np.arange(1, data.shape[-1] + 1)
        self._bins_sum = np.sum(self._bins)
        return data

    def _process(self, data):
        data = np.abs(data)
        yield (data * self._bins).sum(axis=1) / self._bins_sum


class SpectralFlatness(Node):
    """
    Spectral flatness or tonality coefficient, also known as Wiener
    entropy, is a measure used in digital signal processing to characterize an
    audio spectrum. Spectral flatness is typically measured in decibels, and
    provides a way to quantify how tone-like a sound is, as opposed to being
    noise-like. The meaning of tonal in this context is in the sense of the
    amount of peaks or resonant structure in a power spectrum, as opposed to
    flat spectrum of a white noise. A high spectral flatness indicates that
    the spectrum has a similar amount of power in all spectral bands - this
    would sound similar to white noise, and the graph of the spectrum would
    appear relatively flat and smooth. A low spectral flatness indicates that
    the spectral power is concentrated in a relatively small number of
    bands - this would typically sound like a mixture of sine waves, and the
    spectrum would appear "spiky"...

    -- http://en.wikipedia.org/wiki/Spectral_flatness
    """

    def __init__(self, needs=None):
        super(SpectralFlatness, self).__init__(needs=needs)

    def _process(self, data):
        data = np.abs(data)
        mean = data.mean(axis=1)
        mean[mean == 0] = -1e5
        flatness = gmean(data, axis=1) / mean
        yield ArrayWithUnits(flatness, data.dimensions[:1])


class BFCC(Node):
    def __init__(self, needs=None, n_coeffs=13, exclude=1):
        super(BFCC, self).__init__(needs=needs)
        self._n_coeffs = n_coeffs
        self._exclude = exclude

    def _process(self, data):
        data = np.abs(data)
        bfcc = dct(safe_log(data), axis=1) \
            [:, self._exclude: self._exclude + self._n_coeffs]

        yield ArrayWithUnits(
                bfcc.copy(), [data.dimensions[0], IdentityDimension()])
